# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic

from ..objectmeta import ObjectMeta
from .base import BaseNamespace
from .namespacespec import NamespaceSpec


class OrganizationSpec(NamespaceSpec):
    model_config = {'populate_by_name': True}

    id: int | None = pydantic.Field(
        default=None,
        title="Organization ID",
        description=(
            "A numeric identifier for this organization. This field is "
            "auto-generated and can not be set or changed."
        ),
        frozen=True
    )

    display_name: str = pydantic.Field(
        default=...,
        alias='displayName',
        description=(
            "The display name of the `Organization` for user "
            "interfaces."
        )
    )


class Organization(BaseNamespace[ObjectMeta[str]], kind='Namespace', type='webiam.io/organization'):
    model_config = {'populate_by_name': True}

    spec: 'OrganizationSpec' = pydantic.Field(
        default=...,
        description="Defines the behavior of the `Namespace`."
    )


# Copyright (C) 2023-2024 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import inspect
import logging
from typing import Any
from typing import ClassVar
from typing import Generic
from typing import Literal
from typing import Self
from typing import TypeVar

import pydantic
import yaml

from canonical import ResourceName
from canonical import TypedResourceName
from canonical.protocols import IStorage
from canonical.protocols import ITransaction
from .apiresourcelist import APIResourceList
from .listbase import ListBase
from .namespacedobjectmeta import NamespacedObjectMeta
from .objectmeta import ObjectMeta
from .primarykey import PrimaryKey
from .resourcemetadata import ResourceMetadata
from .transientmeta import TransientMeta


__all__: list[str] = [
    'Resource'
]

M = ObjectMetaType = TypeVar('M', bound=ObjectMeta[Any] | TransientMeta)
S = TypeVar('S', bound='Resource[Any]')
V = TypeVar('V')
NOT_PROVIDED: object = object()


class Resource(pydantic.BaseModel, Generic[M]):
    _storage: IStorage[Self] | None = pydantic.PrivateAttr(default=None)
    List: ClassVar[type[ListBase[Any, Any]]]
    metadata_class: ClassVar[type[ObjectMeta[Any] | TransientMeta]]
    model_config = {'populate_by_name': True}
    base_path: ClassVar[str]
    group: ClassVar[str]
    logger: ClassVar[logging.Logger] = logging.getLogger('canonical.ext.resource')
    plural: ClassVar[str]
    KeyType: ClassVar[type[PrimaryKey[type[Self]]]]
    ResourceName: ClassVar[type[TypedResourceName[Any]]]
    __meta__: ClassVar[ResourceMetadata]

    api_version: str = pydantic.Field(
        default=...,
        alias='apiVersion',
        title="API Version",
        description=(
            "The `apiVersion` field defines the versioned schema of this "
            "representation of an object. Servers should convert recognized "
            "schemas to the latest internal value, and may reject "
            "unrecognized values."
        ),
        frozen=True
    )

    kind: str = pydantic.Field(
        default=...,
        description=(
            "Kind is a string value representing the REST resource this "
            "object represents. Servers may infer this from the endpoint "
            "the client submits requests to. Cannot be updated. In `CamelCase`."
        ),
        frozen=True
    )

    metadata: M = pydantic.Field(
        default=...,
        title='Metadata',
        description=(
            "`ObjectMeta` is metadata that all persisted resources "
            "must have, which includes all objects users must create."
        ),
        frozen=True
    )

    @property
    def scoped(self):
        n = self.metadata.name
        if self.is_namespaced():
            assert isinstance(self.metadata, NamespacedObjectMeta)
            n = f'{self.metadata.namespace}/{self.metadata.name}'
        return n

    @classmethod
    def is_namespaced(cls) -> bool:
        metadata_cls = cls.model_fields['metadata'].annotation
        if not inspect.isclass(metadata_cls):
             return False
        assert metadata_cls in {ObjectMeta, NamespacedObjectMeta, TransientMeta}\
            or issubclass(metadata_cls, (ObjectMeta, NamespacedObjectMeta, TransientMeta))
        return metadata_cls.is_namespaced() # type: ignore

    @classmethod
    def is_destroyable(cls) -> bool:
        """Return a boolean indicating if the resource may be destroyed
        by a client.
        """
        return True

    @classmethod
    def is_purgable(cls) -> bool:
        """Return a boolean indicating if the resources may be purged
        by a client.
        """
        return True

    @classmethod
    def new(cls, name: Any, **params: Any):
        return cls.model_validate({
            **params,
            'metadata': {
                'name': name
            }
        })

    @classmethod
    def register(cls, resources: APIResourceList):
        return resources.add(cls.__meta__)

    @property
    def key(self):
        return self.KeyType(
            name=self.metadata.name,
            namespace=getattr(self.metadata, 'name', None)
        )

    @property
    def relname(self) -> str:
        return f'{self.plural}/{self.metadata.name}'

    def __init_subclass__(cls, **kwargs: Any):
        qualname = kwargs.get('version')
        kind = kwargs.get('kind') or cls.__name__
        if qualname:
            name, version = str.split(qualname, '/')
            if '.' in name:
                cls.plural, cls.group = name.split('.', 1)
                api_version = f'{cls.group}/{version}'
            else:
                cls.group = ''
                cls.plural = name
                api_version = version

            base_path = cls.group
            if base_path:
                base_path += '/'
            base_path = f'{base_path}{version}'
            if cls.is_namespaced():
                base_path += '/namespaces/{namespace}'
            cls.base_path = f'{base_path}/{cls.plural}'

            cls.__meta__ = ResourceMetadata(
                api_group=cls.group,
                version=version,
                kind=cls.__name__,
                namespaced=cls.is_namespaced(),
                plural=cls.plural
            )

            # Set defaults and annotations.
            cls.model_fields['api_version'].annotation = Literal[api_version] # type: ignore
            cls.model_fields['api_version'].default = api_version
            cls.model_fields['kind'].annotation = Literal[kind] # type: ignore
            cls.model_fields['kind'].default = kind
            
            # Deprecated
            cls.ResourceName = ResourceName.subclass(cls) # type: ignore

            # Create the primary key type.
            cls.KeyType = PrimaryKey.typed(cls) # type: ignore

            # Create the list type.
            cls.List = type(f'{cls.__name__}List', (ListBase[Literal[f'{cls.__name__}List'], cls],), { # type: ignore
                'items': pydantic.Field(
                    default_factory=list,
                    description=(
                        "The `items` member contains an array "
                        f"of `{cls.__name__}` objects."
                    )
                ),
                '__annotations__': {'items': list[cls]}
            })
            cls.List.model_fields['kind'].default = f'{cls.__name__}List'
            cls.List.model_rebuild()

            metadata_class = cls.model_fields['metadata'].annotation
            if not inspect.isclass(metadata_class):
                return False
            assert metadata_class in {ObjectMeta, NamespacedObjectMeta, TransientMeta}\
                or issubclass(metadata_class, (ObjectMeta, NamespacedObjectMeta, TransientMeta))
            cls.metadata_class = metadata_class
            cls.model_fields['metadata'].default_factory = metadata_class.default # type: ignore
            cls.model_rebuild()

    def get_namespace(self) -> str | None:
        return self.metadata.get_namespace()

    def get_resource_name(self: S, service: str) -> TypedResourceName[S]:
        return self.ResourceName(f'//{service}/{self.plural}/{self.metadata.name}')

    def model_post_init(self, _: Any) -> None:
        self.metadata.attach(self)

    def model_dump_yaml(self, indent: int =2, **kwargs: Any) -> str:
        return yaml.safe_dump(  # type: ignore
            self.model_dump(mode='json', by_alias=True, **kwargs),
            default_flow_style=False
        )

    def get_mutable_data(self) -> dict[str, Any]:
        return self.model_dump(
            mode='json',
            include=self.get_mutable_fields()
        )

    def get_mutable_fields(self) -> set[str]:
        return set()

    def is_created(self):
        return bool(self.metadata.resource_version)

    def is_persistable(self) -> bool:
        return isinstance(self.metadata, (ObjectMeta, NamespacedObjectMeta))

    async def persist(
        self,
        storage: IStorage[Self] | None = None,
        mode: Literal['create', 'replace'] = 'create',
        transaction: ITransaction | None = None
    ) -> Self:
        return await self.persist_key(self.key, storage, mode, transaction)

    async def persist_key(
        self,
        key: Any,
        storage: IStorage[Self] | None = None,
        mode: Literal['create', 'replace'] = 'create',
        transaction: ITransaction | None = None
    ) -> Self:
        storage = storage or self._storage
        if storage is None:
            raise TypeError(
                f'{type(self).__name__} is not attached to any storage '
                'and the `storage` parameter is None.'
            )
        try:
            old = await storage.get(key)
            assert old is not None
        except storage.DoesNotExist:
            old = None
        if mode == 'create' and old is not None:
            raise storage.Duplicate
        if old is not None:
            raise NotImplementedError
        return await storage.persist(self, transaction=transaction, model=key.get_type())

    def replacable(self) -> bool:
        return True


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Protocol


class IDatastoreQuery(Protocol):
    __module__: str = 'cbra.ext.google.types'
    order: list[str]
    def add_filter(self, attname: str, op: str, value: Any) -> None: ...
    def keys_only(self) -> bool: ...
    def fetch(self, *args: Any, **kwargs: Any) -> None: ...


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Generic
from typing import Protocol
from typing import TypeVar

from canonical.exceptions import DoesNotExist
from canonical.exceptions import MultipleObjectsReturned


I = TypeVar('I')


class IRepository(Protocol, Generic[I]):
    __module__: str = 'canonical.protocols'
    InstanceType = I
    DoesNotExist: type[LookupError] = DoesNotExist
    MultipleObjectsReturned: type[LookupError] = MultipleObjectsReturned

    async def delete(self, obj: I | int | str) -> None:
        ...

    async def get(self, pk: Any) -> I | None:
        ...

    async def one(self, pk: Any) -> I:
        ...

    async def persist(self, obj: I) -> None:
        ...


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import functools
from typing import Any
from typing import TypeVar

from pydantic import GetJsonSchemaHandler
from pydantic.json_schema import JsonSchemaValue
from pydantic_core import CoreSchema
from pydantic_core import core_schema


T = TypeVar('T')


class Text(str):

    @classmethod
    def __get_pydantic_core_schema__(cls, *_: Any) -> CoreSchema:
        return core_schema.json_or_python_schema(
            json_schema=core_schema.str_schema(max_length=128),
            python_schema=core_schema.chain_schema([
                core_schema.str_schema(max_length=128),
                core_schema.no_info_plain_validator_function(cls.fromstring)
            ]),
            serialization=core_schema.plain_serializer_function_ser_schema(str)
        )

    @classmethod
    def __get_pydantic_json_schema__(
        cls,
        _: CoreSchema,
        handler: GetJsonSchemaHandler
    ) -> JsonSchemaValue:
        return handler(core_schema.str_schema(max_length=128))

    @classmethod
    def fromstring(cls, v: Any, _: Any = None):
        return cls(
            functools.reduce(
                lambda x, f: f(x),
                [
                    str.strip
                ],
                v
            )
        )

    def __repr__(self) -> str: # pragma: no cover
        return repr(str(self))


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import AsyncIterable
from typing import Generator
from typing import Generic
from typing import TypeVar

import pydantic

from .resource import Resource
from .rootresource import RootResource


R = TypeVar('R', bound=Resource[Any] | RootResource[Any])
K = TypeVar('K')
T = TypeVar('T', bound=pydantic.BaseModel)


class IResourceQuery(AsyncIterable[K], Generic[K]):

    def __await__(self) -> Generator[None, None, list[K]]:
        ...


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

from canonical.ext.resource import Error
from .default import Default


class Put(Default):
    detail = True
    exists = True
    method = 'PUT'
    requires_body = True
    verb = 'replace'

    def get_openapi_responses(self) -> dict[int, Any]:
        return {
            **super().get_openapi_responses(),
            409: {
                'model': Error,
                'description': (
                    f'The {self.model.__name__} identified by the path parameters '
                    'can not be replaced.'
                )
            }
        }


# Copyright (C) 2023-2024 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Generic
from typing import TypeVar

import pydantic


__all__: list[str] = [
    'ResourceKey'
]

T = TypeVar('T')


class ResourceKey(pydantic.BaseModel, Generic[T]):
    api_version: str
    kind: str
    name: T


# Copyright (C) 2023-2024 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import get_args
from typing import Any
from typing import ClassVar
from typing import Generic
from typing import Literal
from typing import Self
from typing import TypeAlias
from typing import TypeVar

import pydantic

from canonical import ResourceName
from canonical import TypedResourceName
from .apiresourcelist import APIResourceList
from .listbase import ListBase
from .primarykey import PrimaryKey
from .resource import Resource
from .resourcemetadata import ResourceMetadata

__all__: list[str] = [
    'RootResource'
]

T = TypeVar('T', bound=Resource[Any])
S = TypeVar('S', bound='RootResource[Any]')


class RootResource(pydantic.RootModel[T], Generic[T]):
    _is_namespaced: bool
    group: ClassVar[str]
    base_path: ClassVar[str]
    plural: ClassVar[str]
    List: ClassVar[type[ListBase[Any, Any]]]
    KeyType: ClassVar[type[PrimaryKey[type[Self]]]]
    ResourceName: ClassVar[type[TypedResourceName[Any]]]
    __meta__: ClassVar[ResourceMetadata]

    @property
    def api_version(self):
        return self.root.api_version

    @property
    def key(self) -> PrimaryKey[type[Self]]:
        return self.KeyType(
            name=self.metadata.name,
            namespace=getattr(self.metadata, 'namespace', None)
        )

    @property
    def kind(self):
        return self.root.kind

    @property
    def metadata(self):
        return self.root.metadata

    @property
    def relname(self):
        return self.root.relname

    @classmethod
    def register(cls, resources: APIResourceList):
        return resources.add(cls.__meta__)

    @classmethod
    def __pydantic_init_subclass__(cls, **kwargs: Any):
        paths: set[str] = set()
        types: tuple[Resource[Any]] = get_args(cls.model_fields['root'].annotation) # type: ignore
        for model in types:
            paths.add(model.base_path)
        if len(paths) > 1:
            raise ValueError(f"All root models must have the same base path.")
        if paths:
            cls.base_path = types[0].base_path
            cls.group = types[0].group
            cls.plural = types[0].plural
            cls._namespaced = types[0].is_namespaced()
            cls.KeyType = PrimaryKey.typed(cls) # type: ignore
            cls.__meta__ = ResourceMetadata.model_validate({
                **types[0].__meta__.model_dump(),
                'kind': cls.__name__
            })
        cls.ResourceName = ResourceName.subclass(cls) # type: ignore
        cls.List = type(f'{cls.__name__}List', (ListBase[Literal[f'{cls.__name__}List'], cls],), {
            'items': pydantic.Field(
                default_factory=list,
                description=(
                    "The `items` member contains an array "
                    f"of `{cls.__name__}` objects."
                )
            ),
            '__annotations__': {'items': list[cls]}
        })
        cls.List.model_fields['kind'].default = f'{cls.__name__}List'
        cls.List.model_rebuild()

        cls.model_config.update({
            'title': cls.__name__
        })
        cls.model_rebuild()

    @classmethod
    def is_destroyable(cls) -> bool:
        return True

    @classmethod
    def is_namespaced(cls) -> bool:
        return cls._is_namespaced

    @classmethod
    def is_purgable(cls) -> bool:
        return True

    def get_namespace(self) -> str | None:
        return self.root.get_namespace()

    def get_resource_name(self: S, service: str) -> TypedResourceName[S]:
        return self.ResourceName(self.root.get_resource_name(service))

    def model_dump_yaml(self, **kwargs: Any):
        return self.root.model_dump_yaml(**kwargs)

    async def persist(self, *args: Any, **kwargs: Any):
        return await self.root.persist_key(self.key, *args, **kwargs)


ResourceType: TypeAlias = Resource[Any] | RootResource[Any]
ResourceTypeVar = TypeVar('ResourceTypeVar', bound=ResourceType)